﻿using System;
using System.IO;
using System.Reflection;
using System.Xml;

namespace MotionDetectPlugin
{
    public static class Config
    {
        /// <summary>
        /// Reads and stores a copy of settings xml file in XmlDocument format.
        /// </summary>
        static XmlDocument settings;

        /// <summary>
        /// Parses the User nodes and stores it for further processing.
        /// </summary>
        static XmlNodeList userNodes;

        /// <summary>
        /// Index of the camera used for detecting motion
        /// </summary>
        public static int CameraIndex { get; }

        /// <summary>
        /// Specifies the inactivity time in minutes after which period the monitor goes to sleep
        /// </summary>
        public static int MonitorSleepTimeInMinutes { get; }

        /// <summary>
        /// Flag to indicate if the monitor can be put to sleep irrespective of users connected(and forgot to disconnect) or not.
        /// </summary>
        public static bool IgnoreConnectedUsersFlag { get; }

        /// <summary>
        /// Flag to indicate if UI needs to be shown 
        /// </summary>
        public static bool ShowUI { get; }

        /// <summary>
        /// Static constructor reads the XML settings file and stores it in a variable.
        /// </summary>
        static Config()
        {
            try
            {
                settings = getXmlSettings();
                if (settings != null)
                {

                    // Get camera index
                    XmlNode node = settings.SelectSingleNode("//Settings/CAMERA_INDEX");
                    if (node != null)
                    {
                        CameraIndex = XmlConvert.ToInt32(node.InnerText);
                    }
                    else
                        CameraIndex = 0;

                    Console.WriteLine("Camera Index: " + CameraIndex);

                    // Read monitor sleep time
                    node = settings.SelectSingleNode("//Settings/MONITOR_SLEEP_TIME");
                    if (node != null)
                    {
                        MonitorSleepTimeInMinutes = XmlConvert.ToInt32(node.InnerText);
                    }
                    else
                        MonitorSleepTimeInMinutes = 5;

                    Console.WriteLine("Monitor Sleep Time: " + MonitorSleepTimeInMinutes);

                    // Read IgnoreConnectedUsersFlag
                    node = settings.SelectSingleNode("//Settings/IGNORE_CONNECTED_USERS");
                    if (node != null)
                    {
                        IgnoreConnectedUsersFlag = XmlConvert.ToBoolean(node.InnerText);
                    }
                    else
                        IgnoreConnectedUsersFlag = true;

                    Console.WriteLine("IgnoreConnectedUsersFlag: " + IgnoreConnectedUsersFlag);

                    // Exempt users

                    userNodes = settings.SelectNodes("//Settings/SHUTDOWN_MONITOR_EXEMPT_USERS/LOGIN");

                    // Read show ui flag
                    node = settings.SelectSingleNode("//Settings/SHOW_UI");
                    if (node != null)
                    {
                        ShowUI = (node.InnerText == "true" ? true: false);
                    }
                    else
                        ShowUI = false;

                    Console.WriteLine("ShowUI: " + ShowUI);
                }

            }
            catch { }

        }

        /// <summary>
        /// Reads and decrypts settings from the settings xml file
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public static string GetAppSetting(string key)
        {
            //XmlDocument settings = getXmlSettings();
            if (settings != null)
            {
                string setting = string.Empty;
                try
                {
                    XmlNode node = settings.SelectSingleNode("//Settings/" + key.ToUpper());
                    setting = node.InnerText;
                    
                }
                catch { return string.Empty; }
                return setting;
            }

            return string.Empty;
        }
        /// <summary>
        /// Returns true if there is a matching node for the key 
        /// </summary>
        /// <param name="key">User login</param>
        /// <returns></returns>
        public static bool IsExemptUser(string key)
        {
            foreach (XmlNode node  in userNodes)
            {
                //Console.WriteLine(node.InnerText);
                if (node.InnerText == key)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Reads the XML settings file, creates XmlDocument object and returns the same.
        /// </summary>
        /// <returns></returns>
        private static XmlDocument getXmlSettings()
        {
            XmlDocument settings = new XmlDocument();
            string path = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
            string filePath = Path.Combine(path, "MotionDetectPluginSettings.xml");
            try { settings.Load(filePath); }
            catch { }
            return settings;
        }
    }
}
