﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Intel.CFC.Plugin;
using System.Threading;
using OpenCvSharp;
using System.Diagnostics;

namespace MotionDetectPlugin
{
    public enum MOTION_STATUS
    {
        UNKNOWN = -1,
        NO_MOTION = 0,
        MOTION = 1,
    };

    /// <summary>
    /// This plugin is used to control display state(on / standby) depending on user motion detected by camera.
    /// </summary>
    public class MotionDetectPlugin : CFCPlugin
    {
        PluginUI UI = null;
        PluginInfo pluginDetails = new PluginInfo();
        PluginUIElementGroup motionGroup = new PluginUIElementGroup();

        string HubText = "";
        const string MOTION = "00000000-0000-0000-0000-000000000008";
        const string NO_MOTION = "00000000-0000-0000-0000-000000000009";
        static bool isRunning = true;
        static bool isMonitorOff = false;
        private int _configuredSleepTimeInSeconds   = Config.MonitorSleepTimeInMinutes * 60;
        private bool _ignoreConnectedUsersFlag      = Config.IgnoreConnectedUsersFlag;
        private bool _showUI                        = Config.ShowUI;
        private int _numUsersConnected = 0;
        private bool _checkDisplaySleepMsgTimerFlag = true;
        private Stopwatch _lastActivityTimer = new Stopwatch();
        private byte[] offImageBytes = ResourceToBytes(new Uri("/MotionDetectPlugin;component/images/Off.png", System.UriKind.Relative));
        private byte[] onImageBytes = ResourceToBytes(new Uri("/MotionDetectPlugin;component/images/On.png", System.UriKind.Relative));
        private byte[] unknownStateImageBytes = ResourceToBytes(new Uri("/MotionDetectPlugin;component/images/camera.png", System.UriKind.Relative));
        private int _cameraIndex = Config.CameraIndex;

        /// <summary>
        /// Displays appropriate icon to show whether Camera has detected any motion or not. 
        /// </summary>
        /// <param name="state">Indicates Motion or no motion</param>
        public void ShowMotionStatusPluginUI(MOTION_STATUS state)
        {
            Console.WriteLine("ShowUI: " + _showUI + "state(0 = no motion, 1 = motion): " + state.ToString());

            if (!_showUI)
                return;

            if (state == MOTION_STATUS.NO_MOTION)
            {
                motionGroup.GroupName = "No motion";
                motionGroup.ImageBytes = offImageBytes;
            }
            else if (state == MOTION_STATUS.MOTION)
            {
                motionGroup.GroupName = "Motion";
                motionGroup.ImageBytes = onImageBytes;
            }
            else if (state == MOTION_STATUS.UNKNOWN)
            {
                motionGroup.GroupName = "";
                motionGroup.ImageBytes = unknownStateImageBytes;
            }

            FireUIUpdated();
            FireHubTextUpdated();
        }

        /// <summary>
        /// Starts a thread to detect motion.
        /// </summary>
        public void InitPlugin()
        {
            // Create a thread to detect motion
            System.Threading.ThreadStart ts = new ThreadStart(this.DetectMotion);
            Thread t = new Thread(ts);
            t.Start();
     
        }


        /// <summary>
        /// This method uses Open Source Computer Vision(OpenCV) Library to detect motion using video camera input stream
        /// </summary>
        public void DetectMotion()
        {

            // How much change is needed for a change to be detected. 0 - 255, larger = more change
            double thresdhold = 16;

            // How much of the screen must change to assume there was motion. .00 - 1.00.
            double minPercentChangeToTriggerMotion = .05;

            // How much change is too much to assume there was motion - filter out light changes or camera movement. .00 - 1.00.
            double maxPercentChangeToTriggerMotion = .5;

            // How many frames (out of 10) must change for motion to be triggerd. Stops a quick blip of motion (like lights turning off) from triggering.
            int motionResistance = 5; 

            VideoCapture vc = null;
            double numberOfPixels;

            try
            {
                vc = VideoCapture.FromCamera(_cameraIndex);
                int ultraHighRes = 10000;
                vc.Set(CaptureProperty.FrameHeight, ultraHighRes);
                vc.Set(CaptureProperty.FrameWidth, ultraHighRes);
                double width = vc.Get(CaptureProperty.FrameWidth);
                double height = vc.Get(CaptureProperty.FrameHeight);
                vc.Grab();
                DebugOutput("Camera resolution");
                DebugOutput("Width: " + width.ToString());
                DebugOutput("Height: " + height.ToString());
                numberOfPixels = width * height;

            }
            catch (Exception)
            {
                DebugOutput("Can't connect to the camera");
                isRunning = false;
                return;
            }

            // init vars
            Mat currentFrame = null;
            Mat lastFrame = null;
            Mat diffFrame = new Mat();
            List<int> isTriggered = new List<int>();
            MOTION_STATUS prevState = MOTION_STATUS.UNKNOWN;
            try
            {
                while (isRunning)
                {
                    vc.Grab(); // grab a frame
                    currentFrame = vc.RetrieveMat(); // get the frame data into a var
                                                     // Convert frame to Gray
                    Cv2.CvtColor(currentFrame, currentFrame, ColorConversionCodes.BGR2GRAY);

                    // Apply Gaussian Blur
                    Cv2.GaussianBlur(currentFrame, currentFrame, new Size(15, 15), 0);

                    // Motion Detection
                    if (lastFrame != null)
                    {
                        // Get the differences between the 2 frames
                        Cv2.Absdiff(currentFrame, lastFrame, diffFrame);

                        // Remove everything that's below the threshold
                        Cv2.Threshold(diffFrame, diffFrame, thresdhold, 255, ThresholdTypes.Binary);

                        // How many pixels changed?
                        double pixelChangeCount = Cv2.CountNonZero(diffFrame);

                        // If more than the desired percentage changed, record it as a change
                        double percentChange = pixelChangeCount / numberOfPixels;
                        //DebugOutput("Percent Change: " + percentChange);

                        // Check if the percentage of change is within range for motion.
                        if (percentChange >= minPercentChangeToTriggerMotion && percentChange <= maxPercentChangeToTriggerMotion)
                        {
                            //DebugOutput("Percent in Range for motion");
                            isTriggered.Add(1);
                        }
                        else
                        {
                            //DebugOutput("Percent out of range for motion");
                            isTriggered.Add(0);
                        }
                        if (isTriggered.Count > 10)
                        {
                            isTriggered.RemoveAt(0);
                        }

                        // Calculate Resistance
                        //DebugOutput("Resistance: " + isTriggered.Sum());
                        if (isTriggered.Sum() >= motionResistance)
                        {
                            // Reset the last activity time
                            _lastActivityTimer = Stopwatch.StartNew();
                            // Reset the flag to enable showing toast message for display 
                            _checkDisplaySleepMsgTimerFlag = true;

                            // Display active icon
                            // Update Plugin Group UI
                            if (prevState != MOTION_STATUS.MOTION)
                            {
                                ShowMotionStatusPluginUI(MOTION_STATUS.MOTION);
                                prevState = MOTION_STATUS.MOTION;
                            }

                            //Console.WriteLine("MOTION");
                            if (isMonitorOff)
                            {
                                MonitorControl.Wake();
                                isMonitorOff = false;                                
                            }
                        }
                        else
                        {
                            //Console.WriteLine("NO MOTION");
                            // Display inactive icon
                            // Update Plugin Group UI
                            if (prevState != MOTION_STATUS.NO_MOTION)
                            {
                                ShowMotionStatusPluginUI(MOTION_STATUS.NO_MOTION);
                                prevState = MOTION_STATUS.NO_MOTION;
                            }

                            if (!IsExemptUserLoggedIn() && (_ignoreConnectedUsersFlag || (!_ignoreConnectedUsersFlag && (_numUsersConnected <= 0))))
                            { // OK to switch off the monitor irrespective of connected users
                                TimeSpan ts = _lastActivityTimer.Elapsed;

                                if(_checkDisplaySleepMsgTimerFlag && 
                                   ((_configuredSleepTimeInSeconds - ts.TotalSeconds) <= 30 ))
                                {
                                    ShowHubToast("Monitor will go to standby mode in 30 seconds", null , 2);
                                    _checkDisplaySleepMsgTimerFlag = false;
                                }

                                if (ts.TotalSeconds >= _configuredSleepTimeInSeconds)
                                {
                                    //MonitorControl.Sleep();
                                    MonitorControl.Sleep();
                                    isMonitorOff = true;
                                }
                            }
                        }
#if DEBUG
                        //Cv2.ImShow("Motion", diffFrame); // Comment motion to hide the window
                        Cv2.WaitKey(1);                  // Comment motion to hide the window
#endif
                        lastFrame.Dispose();
                    }
                    //DebugOutput(" ");
                    lastFrame = currentFrame.Clone();
                    currentFrame.Dispose();
                }
            }
            //catch (ThreadAbortException)
            //{
            //    DebugOutput("Detection Thread Aborted Unexpectedly.");
            //}
            catch (OpenCVException ocve)
            {
                DebugOutput("Open CV Exception: " + ocve.ToString());
            }
            catch (Exception e)
            {
                DebugOutput("Exception: " + e.ToString());
            }
            isRunning = false;
            if (vc != null)
            {
                vc.Release();
            }
            try
            {
                Cv2.DestroyAllWindows();
            }
            catch
            {

            }
        }
        /// <summary>
        /// Used to find out if a specific user has logged on.
        /// </summary>
        /// <returns></returns>
        public bool IsExemptUserLoggedIn()
        {
            // Get current user and find if the user is exempt
            bool result = Config.IsExemptUser(System.Environment.UserName);

            //Console.WriteLine("IsExemptUserLoggedIn returned: " + result);

            return result;
        }

        static void DebugOutput(string message)
        {
#if DEBUG
            Console.WriteLine(message);
#endif
        }

        /// <summary>
        /// This method is called when the plugin is actually loaded.
        /// </summary>
        public override void Load()
        {
            //Do nothing
            LogMessage("Motion Detect Plugin Load entered", null);
            Console.WriteLine("Motion Detect Plugin Load entered");
            Console.WriteLine("Camera Index: " + _cameraIndex.ToString());
            Console.WriteLine("Configured sleep time: " + _configuredSleepTimeInSeconds.ToString());
            Console.WriteLine("IgnoreConnectedUsersFlag: " + _ignoreConnectedUsersFlag.ToString());            
            Console.WriteLine("ShowUI: " + _showUI.ToString());

            // Set the elapsed time to zero and starts measuring elapsed time
            _lastActivityTimer = Stopwatch.StartNew();

            if (_showUI)
            {
                pluginDetails.Name = "Motion Detect Plugin";
                pluginDetails.Id = new Guid("37fe3db7-18eb-41c8-af9d-c670675432fe");
                pluginDetails.Description = "Motion Detect Plugin";
                pluginDetails.Copyright = "Copyright (C) 2017 Intel Corporation - All Rights Reserved";
                pluginDetails.Company = "Intel Corporation";
                UI = new PluginUI();
                UI.pluginInfo = pluginDetails;
                UI.Groups = new List<PluginUIElementGroup>();
                motionGroup.GroupName = "Check Motion";

                motionGroup.ImageBytes = ResourceToBytes(new Uri("/MotionDetectPlugin;component/images/camera.png",
                                                         System.UriKind.Relative));
                UI.Groups.Add(motionGroup);

            }

            InitPlugin();

            LogMessage("Motion Detect Plugin Load leaving.", null);
            Console.WriteLine("Motion Detect Plugin Load leaving.");

        }
        public override void UnLoad()
        {
            //Do nothing
            LogMessage("Plugin Unloaded", null);
        }
        public override void UserConnected(UserEventArgs e)
        {
            _numUsersConnected++;
            LogMessage("Plugin User Connect", null);
        }
        public override void UserDisconnected(UserEventArgs e)
        {
            _numUsersConnected--;
            LogMessage("Plugin User Disconnect Loaded", null);
        }
        public override void UserPresentationStart(UserEventArgs e)
        {
            //Do nothing
            LogMessage("Plugin Presentation Started", null);
        }
        public override void UserPresentationEnd(UserEventArgs e)
        {
            //Do nothing
            LogMessage("Plugin Presentation End", null);
        }
        public override void UIElementEvent(UIEventArgs e)
        {
            //Do nothing
            LogMessage("Plugin Received UI Event: " + e.ElementId.ToString(), null);
            //FireUIUpdated();
        }
        public override PluginUI GetUI(UserEventArgs e)
        {
            return UI;
        }
        public override PluginInfo GetPluginInfo()
        {
            return pluginDetails;
        }
        public override string GetHubText()
        {
            return HubText;
        }
    }
}